# ©2023, ANSYS Inc. Unauthorized use, distribution or duplication is prohibited.

"""Frontend of the node status overview view."""

from dash_extensions.enrich import html
import optislang_dash_lib


def layout(project_status_info: dict) -> html.Div:
    """Layout of the status overview view."""

    if project_status_info:
        return html.Div(
            id="node-status-view",
            children=[
                optislang_dash_lib.Nodestatusviewcomponent(
                    id="node-status-view-component",
                    project_state=project_status_info,
                ),
            ],
        )
    else:
        return html.Div([])



# ©2023, ANSYS Inc. Unauthorized use, distribution or duplication is prohibited.

"""Backend of the monitoring step."""

from ansys.saf.glow.solution import StepModel


class MonitoringStep(StepModel):
    """Step model of the monitoring step."""



# ©2023, ANSYS Inc. Unauthorized use, distribution or duplication is prohibited.

"""Backend of the problem setup step."""

import json
from pathlib import Path
import platform
import time
from typing import List
import subprocess
import sys

from ansys.optislang.core import Optislang, logging
from ansys.saf.glow.solution import FileReference, AssetFileReference, StepModel, StepSpec, long_running, transaction
from ansys.solutions.optislang.frontend_components.project_properties import ProjectProperties, write_properties_file, apply_placeholders_to_properties_file
from ansys.solutions.products_ecosystem.controller import AnsysProductsEcosystemController
from ansys.solutions.products_ecosystem.utils import convert_to_long_version

from ansys.solutions.{{ cookiecutter.__solution_name_slug }}.model.osl_project_tree import dump_project_state, get_project_tree, get_node_list, get_step_list
from ansys.solutions.{{ cookiecutter.__solution_name_slug }}.ui.utils.monitoring import _get_actor_hids, read_optislang_logs


class ProblemSetupStep(StepModel):
    """Step model of the problem setup step."""

    # Parameters ------------------------------------------------------------------------------------------------------

    # Frontend persistence
    ansys_ecosystem_ready: bool = False
    optislang_solve_status: str = "initial"  # initial, processing, finished, stopped, aborted, idle
    ui_placeholders: dict = {}
    app_metadata: dict = {}
    analysis_running: bool = False

    # Backend data model
    tcp_server_host: str = "127.0.0.1"
    tcp_server_port: int = None
    ansys_ecosystem: dict = {
        "optislang": {
            "authorized_versions": ["2022.2", "2023.1"],
            "installed_versions": [],
            "compatible_versions": [],
            "selected_version": None,
            "alert_message": "OptiSLang install not checked.",
            "alert_color": "warning",
            "alias": "optiSLang",
        }
    }
    step_list: list = []
    node_list: list = []
    placeholders: dict = {}
    registered_files: List = []
    settings: dict = {}
    parameter_manager: dict = {}
    criteria: dict = {}
    project_status_info: dict = {}
    actors_info: dict = {}
    actors_status_info: dict = {}
    results_files: dict = {}
    tcp_server_stopped_states = ["idle", "finished", "stopped", "aborted"]
    optislang_logs: list = []
    optislang_log_level: str = "DEBUG"
    project_initialized: bool = False
    has_project_state: bool = False

    # File storage ----------------------------------------------------------------------------------------------------

    # Inputs
    project_file: FileReference = FileReference("Problem_Setup/{{ cookiecutter.__optiSLang_application_archive_stem }}.opf")
    properties_file: FileReference = FileReference("Problem_Setup/{{ cookiecutter.__optiSLang_application_archive_stem }}.json")
    metadata_file: FileReference = FileReference("Problem_Setup/metadata.json")
    project_state_file: FileReference = FileReference("Problem_Setup/project_state.json")

    # Outputs
    working_properties_file: FileReference = FileReference("Problem_Setup/working_properties_file.json")
    server_info_file: FileReference = FileReference("Problem_Setup/server_info.ini")
    optislang_log_file: FileReference = FileReference("Problem_Setup/pyoptislang.log")

    # Methods ---------------------------------------------------------------------------------------------------------

    @transaction(
        self=StepSpec(
            upload=["has_project_state"]
        )
    )
    def generate_project_state(self) -> None:
        """Generate a project state from an optiSLang opf file."""

        project_file = Path(__file__).absolute().parent.parent / "model" / "assets" / "{{ cookiecutter.__optiSLang_application_archive_stem }}.opf"

        dump_project_state(project_file, Path(project_file).parent / "project_state.json")

        self.has_project_state = True

    @transaction(
        self=StepSpec(
            download=["project_state_file"],
            upload=["step_list", "node_list"]
        )
    )
    def read_project_tree(self) -> None:
        """Read project tree from optiSLang project state file."""

        project_tree = get_project_tree(self.project_state_file.path)
        self.step_list = get_step_list(project_tree)
        self.node_list = get_node_list(project_tree)

    @transaction(
        self=StepSpec(
            upload=[
                "project_file",
                "properties_file",
                "metadata_file",
                "project_state_file",
            ]
        )
    )
    def upload_bulk_files_to_project_directory(self) -> None:
        """Upload bulk files to project directory."""

        original_project_file = Path(__file__).parent.absolute().parent / "model" / "assets" / "{{ cookiecutter.__optiSLang_application_archive_stem }}.opf"
        self.project_file.write_bytes(original_project_file.read_bytes())

        original_properties_file = (
            Path(__file__).parent.absolute().parent / "model" / "assets" / "{{ cookiecutter.__optiSLang_application_archive_stem }}.json"
        )
        self.properties_file.write_bytes(original_properties_file.read_bytes())

        original_metadata_file = Path(__file__).parent.absolute().parent / "model" / "assets" / "metadata.json"
        self.metadata_file.write_bytes(original_metadata_file.read_bytes())

        original_project_state_file = Path(__file__).parent.absolute().parent / "model" / "assets" / "project_state.json"
        self.project_state_file.write_bytes(original_project_state_file.read_bytes())

    @transaction(
        self=StepSpec(
            upload=["ansys_ecosystem", "ansys_ecosystem_ready"],
        )
    )
    def check_ansys_ecosystem(self) -> None:
        """Check if Ansys Products are installed and if the appropriate versions are available."""

        self.ansys_ecosystem_ready = True

        controller = AnsysProductsEcosystemController()

        for product_name in self.ansys_ecosystem.keys():

            self.ansys_ecosystem[product_name]["installed_versions"] = controller.get_installed_versions(
                product_name, outout_format="long"
            )
            self.ansys_ecosystem[product_name]["compatible_versions"] = [
                product_version
                for product_version in self.ansys_ecosystem[product_name]["installed_versions"]
                if product_version in self.ansys_ecosystem[product_name]["authorized_versions"]
            ]

            if len(self.ansys_ecosystem[product_name]["installed_versions"]) == 0:
                alert_message = f"No installation of {product_name.title()} found in the machine {platform.node()}."
                alert_color = "danger"
                self.ansys_ecosystem_ready = False
            elif len(self.ansys_ecosystem[product_name]["compatible_versions"]) == 0:
                alert_message = (
                    f"None of the authorized versions of {product_name.title()} "
                    f"is installed in the machine {platform.node()}.\n"
                )
                alert_message += "At least one of these versions is required:"
                for authorized_version in self.ansys_ecosystem[product_name]["authorized_versions"]:
                    self.ansys_ecosystem[product_name][
                        "alert_message"
                    ] += f" {convert_to_long_version(authorized_version)}"
                alert_message += "."
                alert_color = "danger"
                self.ansys_ecosystem_ready = False
            else:
                self.ansys_ecosystem[product_name]["selected_version"] = self.ansys_ecosystem[product_name][
                    "compatible_versions"
                ][
                    -1
                ]  # Latest
                alert_message = f"{product_name.title()} install detected. Compatible versions are:"
                for compatible_version in self.ansys_ecosystem[product_name]["compatible_versions"]:
                    alert_message += f" {convert_to_long_version(compatible_version)}"
                alert_message += ".\n"
                alert_message += "Selected version is %s." % (self.ansys_ecosystem[product_name]["selected_version"])
                alert_color = "success"
            self.ansys_ecosystem[product_name]["alert_message"] = alert_message
            self.ansys_ecosystem[product_name]["alert_color"] = alert_color

    @transaction(self=StepSpec(download=["properties_file"], upload=["placeholders", "registered_files", "settings", "parameter_manager", "criteria"]))
    def get_default_placeholder_values(self):
        """Get placeholder values and definitions using the ProjectProperties class."""

        pp = ProjectProperties()
        pp.read_file(self.properties_file.path)
        self.placeholders = pp._placeholders
        self.registered_files = pp._registered_files
        self.settings = pp._settings
        self.parameter_manager = pp._parameter_manager
        self.criteria = pp._criteria

    @transaction(self=StepSpec(download=["properties_file", "ui_placeholders"], upload=["working_properties_file"]))
    def write_updated_properties_file(self) -> None:
        properties = apply_placeholders_to_properties_file(self.ui_placeholders, self.properties_file.path)
        write_properties_file(properties, Path(self.working_properties_file.path))

    @transaction(
        self=StepSpec(
            download=["metadata_file"],
            upload=["app_metadata"]
        )
    )
    def get_app_metadata(self) -> None:
        """Read OWA metadata file."""

        with open(self.metadata_file.path) as f:
            self.app_metadata = json.load(f)

    @transaction(
        self=StepSpec(
            download=[
                "project_file",
                "working_properties_file",
                "tcp_server_stopped_states",
                "optislang_log_level",
                "node_list",
            ],
            upload=[
                "optislang_solve_status",
                "server_info_file",
                "actors_info",
                "actors_status_info",
                "tcp_server_port",
                "project_status_info",
                "results_files",
                "optislang_log_file",
                "optislang_logs",
            ],
        )
    )
    @long_running
    def start_analysis(self) -> None:
        """Start optiSLang and run the project."""

        osl_logger = logging.OslLogger(
            loglevel=self.optislang_log_level,
            log_to_file=True,
            logfile_name=self.optislang_log_file.path,
            log_to_stdout=True,
        )

        osl = Optislang(
            project_path=self.project_file.path,
            loglevel=self.optislang_log_level,
            reset=True,
            shutdown_on_finished=True,
            import_project_properties_file=self.working_properties_file.path,
            additional_args=[f"--write-server-info={self.server_info_file.path}"],
            ini_timeout=50,  # might need to be adjusted
        )

        osl.__logger = osl_logger.add_instance_logger(osl.name, osl, self.optislang_log_level)

        if self.tcp_server_port is None:
            if self.server_info_file.exists():
                with open(self.server_info_file.path, "r") as file:
                    lines = [line.rstrip("\n") for line in file.readlines()]
                for line in lines:
                    if line.startswith("server_port="):
                        self.tcp_server_port = int(line.split("=")[1])
                        break
            else:
                raise Exception("No server info file detected. Unable to retrieve TCP port number.")

        osl.start(wait_for_started=True, wait_for_finished=False)

        while True:
            # Get project status info
            self.project_status_info = osl.get_osl_server().get_full_project_status_info()
            # Read pyoptislang logs
            self.optislang_logs = read_optislang_logs(self.optislang_log_file.path)
            # Get actor status info
            for node_info in self.node_list:
                self.actors_info[node_info["uid"]] = osl.get_osl_server().get_actor_info(node_info["uid"])
                node_hids = _get_actor_hids(osl.get_osl_server().get_actor_states(node_info["uid"]))
                if len(node_hids):
                    self.actors_status_info[node_info["uid"]] = []
                    for hid in node_hids:
                        self.actors_status_info[node_info["uid"]].append(
                            osl.get_osl_server().get_actor_status_info(node_info["uid"], hid)
                        )
            # Get status
            self.optislang_solve_status = osl.project.get_status().lower()
            osl.log.info(f"Analysis status: {self.optislang_solve_status}")
            # Upload fields
            self.transaction.upload(["optislang_solve_status"])
            self.transaction.upload(["project_status_info"])
            self.transaction.upload(["actors_info"])
            self.transaction.upload(["actors_status_info"])
            self.transaction.upload(["optislang_logs"])
            # Check if analysis stopped
            if self.optislang_solve_status in self.tcp_server_stopped_states:
                break
            time.sleep(3)

        osl.dispose()



# ©2023, ANSYS Inc. Unauthorized use, distribution or duplication is prohibited.

"""Provide a component to handle the actor statistics for system and actor node types."""

from dash import dash_table
import pandas as pd

from ansys.solutions.{{ cookiecutter.__solution_name_slug }}.ui.utils.common_functions import convert_microseconds


class ActorStatisticsTable:
    """Actor statistics component."""

    def __init__(self) -> None:
        """Constructor."""

        self.actor_info: dict = None
        self.font_size: str = "15px"

    def _get_data(self) -> pd.DataFrame:

        actor_statistics_data = {
            "row_names": ["Usages", "Accumulated", "Minimum", "Maximum", "Mean", "std_dev"],
            "Current Run": [None, None, None, None, None, None],
            "All Runs": [None, None, None, None, None, None],
        }

        if self.actor_info:
            for column_name in ["Current Run", "All Runs"]:
                key = column_name.split()[0].lower()
                actor_statistics_data[column_name][0] = self.actor_info["usage_stats"][key]["num_usages"]
                for index, row_name in enumerate(actor_statistics_data["row_names"]):
                    if row_name != "Usages":
                        duration = self.actor_info["usage_stats"][key]["exec_duration_us"][row_name.lower()]
                        duration = convert_microseconds(duration)
                        actor_statistics_data[column_name][index] = duration

        return pd.DataFrame(actor_statistics_data)

    def render(self):
        """Generate table."""

        data = self._get_data()

        return dash_table.DataTable(
            data=data.to_dict("records"),
            columns=[
                {"name": i, "id": i, "type": "text"} if i != "row_names" else {"name": "", "id": i, "type": "text"}
                for i in data.columns
            ],
            fixed_rows={"headers": True},
            style_header={
                "textAlign": "left",
                "font_family": "Roboto",
                "font_size": self.font_size,
                "fontWeight": "bold",
            },
            style_cell={
                "textAlign": "left",
                "font_family": "Roboto",
                "font_size": self.font_size,
            },
            style_cell_conditional=[
                {"if": {"column_id": "row_names"}, "minWidth": "50px", "maxWidth": "50px", "width": "50px"},
                {"if": {"column_id": "Current Run"}, "minWidth": "50px", "maxWidth": "50px", "width": "50px"},
                {"if": {"column_id": "All Runs"}, "minWidth": "50px", "maxWidth": "50px", "width": "50px"},
            ],
            style_data_conditional=[
                {
                    "if": {"column_id": "row_names", "filter_query": '{Level} eq "std_dev"'},
                    "backgroundColor": "rgb(227, 245, 252)",
                    "color": "rgb(0, 0, 0)",
                    "textAlign": "center",
                    # "text": dcc.Markdown('$Area (m^{2})$', mathjax=True),
                }
            ],
            style_as_list_view=True,
            # markdown_options={"html": True},
        )



# ©2023, ANSYS Inc. Unauthorized use, distribution or duplication is prohibited.

"""Provide a component to handle the root project information."""

from dash import dash_table
import pandas as pd


class ProjectSummaryTable:
    """Project summary component."""

    def __init__(self) -> None:
        """Constructor."""

        self.project_status_info: dict = None
        self.font_size: str = "15px"

    def _get_data(self) -> pd.DataFrame:

        project_summary_data = {
            "column_a": [
                "State",
                "Id",
                "Name",
                "Machine",
                "Location",
                "Project directory",
                "Owner",
                "Registered",
                "Lock info",
            ],
            "column_b": [
                None,
                None,
                None,
                None,
                None,
                None,
                None,
                None,
                None,
            ],
        }

        if self.project_status_info:
            project_summary_data["column_b"] = [
                self.project_status_info["projects"][0]["state"],
                self.project_status_info["projects"][0]["project_id"],
                self.project_status_info["projects"][0]["name"],
                self.project_status_info["projects"][0]["machine"],
                self.project_status_info["projects"][0]["location"],
                self.project_status_info["projects"][0]["working_dir"],
                self.project_status_info["projects"][0]["user"],
                "",
                "",
            ]

        return pd.DataFrame(project_summary_data)

    def render(self):
        """Generate table."""

        data = self._get_data()

        return dash_table.DataTable(
            data=data.to_dict("records"),
            columns=[{"name": i, "id": i, "type": "text"} for i in data.columns],
            fixed_rows={"headers": True},
            style_header={
                "textAlign": "left",
                "font_family": "Roboto",
                "font_size": self.font_size,
                "fontWeight": "bold",
                "border": "none",
                "display": "none",
            },
            style_cell={"textAlign": "left", "font_family": "Roboto", "font_size": self.font_size, "border": "none"},
            style_as_list_view=True,
            style_table={"border": "none"},  # Hide table borders
        )



# ©2023, ANSYS Inc. Unauthorized use, distribution or duplication is prohibited.

"""Frontend of the design table view."""

import dash_bootstrap_components as dbc
from dash_extensions.enrich import html

from ansys.solutions.{{ cookiecutter.__solution_name_slug }}.ui.components.design_table import DesignTable


def layout(actors_info: list, actors_status_info: dict, uid: str, font_size: str = "15px") -> html.Div:
    """Layout of the design table view."""

    design_table = DesignTable()

    if uid in actors_info.keys():
        design_table.actor_info = actors_info[uid]
    if uid in actors_status_info.keys():
        design_table.actor_status_info = actors_status_info[uid][
            0
        ]  # the index 0 here means that we arbitrarily select the 1st hid

    return html.Div(
        [
            html.Br(),
            dbc.Row(
                [
                    dbc.Col(design_table.render(), width=12),
                ]
            ),
        ]
    )



# ©2023, ANSYS Inc. Unauthorized use, distribution or duplication is prohibited.

"""Frontend of the visualization view."""

from dash_extensions.enrich import html
import optislang_dash_lib


def layout(project_status_info: dict, uid: str) -> html.Div:
    """Layout of the visualization view."""

    return html.Div(
        [
            optislang_dash_lib.VisualizationComponent(
                id="visualization-component",
                project_state=project_status_info,
                system_id=uid,
                state_idx=0,
            ),
        ]
    )



# ©2023, ANSYS Inc. Unauthorized use, distribution or duplication is prohibited.

import dash_bootstrap_components as dbc
from dash_extensions.enrich import html

from ansys.solutions.{{ cookiecutter.__solution_name_slug }}.solution.problem_setup_step import ProblemSetupStep


def update_alerts(problem_setup_step: ProblemSetupStep) -> list:
    """Update all Alerts."""

    alerts = []

    # Product version alerts
    for product_name, product_data in problem_setup_step.ansys_ecosystem.items():
        alerts.append(
            html.Div(
                [
                    dbc.Button(
                        f"{product_data['alias']} Version",
                        id=f"popover_{product_name}_version_target",
                        disabled=False,
                        color=product_data["alert_color"],
                        n_clicks=0,
                    ),
                    dbc.Popover(
                        [
                            dbc.PopoverBody(product_data["alert_message"]),
                        ],
                        id=f"popover_{product_name}_version",
                        target=f"popover_{product_name}_version_target",
                        placement="top",
                        is_open=False,
                    ),
                ]
            ),
        )

    # optiSLang solve alert
    if problem_setup_step.optislang_solve_status == "initial":
        solve_message, solve_color = "optiSLang simulation not started.", "warning"
    elif problem_setup_step.optislang_solve_status == "processing":
        solve_message, solve_color = "optiSLang simulation in progress.", "primary"
    elif problem_setup_step.optislang_solve_status == "stopped":
        solve_message, solve_color = "optiSLang simulation stopped.", "danger"
    elif problem_setup_step.optislang_solve_status == "aborted":
        solve_message, solve_color = "optiSLang simulation aborted.", "danger"
    elif problem_setup_step.optislang_solve_status == "finished":
        solve_message, solve_color = "optiSLang simulation completed successfully.", "success"
    else:
        raise ValueError(f"Unknown optiSLang status: {problem_setup_step.optislang_solve_status}.")

    alerts.append(
        html.Div(
            [
                dbc.Button(
                    "optiSLang Solve",
                    id="popover_optislang_solve_target",
                    disabled=False,
                    color=solve_color,
                    n_clicks=0,
                ),
                dbc.Popover(
                    [
                        dbc.PopoverBody(solve_message),
                    ],
                    id="popover_optislang_solve",
                    target="popover_optislang_solve_target",
                    placement="top",
                    is_open=False,
                ),
            ]
        ),
    )

    return alerts


def update_monitoring_alert(problem_setup_step: ProblemSetupStep) -> dbc.Alert:

    if problem_setup_step.optislang_solve_status == "initial":
        return dbc.Alert(
            "No analysis started. No data to display.",
            color="warning",
        )



# ©2023, ANSYS Inc. Unauthorized use, distribution or duplication is prohibited.
# CAPTURED v0.5.1 on 7/4/23 from:
# https://github.com/Solution-Applications/common-files/blob/v0.4.0/setup-environment/setup_environment.py

"""
A Python script to automate the setup of the Python ecosystem of a project.

Prerequisites
-------------

1. The following packages are required:
    * packaging
    * toml

2. This script needs to be executed at project's root.

3. Currently, this code only supports:
    * ``windows`` as operating system
    * ``poetry`` as dependency management system
    * projects without dependency management systems
    * ``poetry`` 1.2 to latest.

4. The following project structure is expected for projects without a dependency management system:
    project-name
    ├──requirements/                    # Folder containing the optional group of dependencies.
    │  ├── requirements_doc.txt         # Requirements file associated to the documentation group.
    │  ├── requirements_tests.txt       # Requirements file associated to the tests group.
    │  ├── requirements_style.txt       # Requirements file associated to the style group.
    │  └── requirements_build.txt       # Requirements file associated to the build group.

5. The following project structure is expected for projects with a dependency management system:
    project-name
    └── pyproject.toml                  # Configuration of the build system.

The following packages are needed on top of the standard Python configuration: ``toml`` and ``packaging``.

Usage
-----

To create a virtual environment and install the dependency management system of the project (if any):
    ``python setup_environment.py``

To install a particular dependency group use the ``-d`` option:
    ``python setup_environment.py -d run`` installs production dependencies
    ``python setup_environment.py -d doc`` installs doc dependencies
    ``python setup_environment.py -d tests`` installs tests dependencies
    ``python setup_environment.py -d style`` installs style dependencies
    ``python setup_environment.py -d build`` installs build dependencies

It is possible to combine several groups:
    ``python setup_environment.py -d run doc tests``

To install all available dependencies (production and optional) use the ``all`` option:
    ``python setup_environment.py -d all``

Extra dependency groups refer to any group declared in the configuration file of the dependency management system which
are not part of doc, tests, style and build. To install these groups use the ``x`` option:
    ``python setup_environment.py -x <name-of-the-group>``

There are two locations where the script will search for optional dependencies:
    * First it checks the ``pyproject.toml`` configuration file and search for optional dependency groups
    * Alternatively, it looks for a ``requirements`` folder at project root containing requirements files with the
      name of the dependency group. For instance: requirements/requirements_doc.txt for the documentation group.

To enforce the version of the dependency manager two options are possible:

    1. ``setup_environment.py`` reads the ``pyproject.toml`` and looks for a ``build-system-version`` key within
    the ``build-system-requirements`` section. The example below illustrates this. Note that,
    ``build-system-requirements` and ``build-system-version`` are not part of the ``PEP518`` standard. These syntax is
    purely internal.
    ```
    [build-system-requirements]
    build-system-version = "1.4.2"
    ```

    2. If the ``pyproject.toml`` does not contain a ``build-system-version``, the ``setup_environment.py`` checks the
    ``-s`` option which can be provided in the command line when executing the script. In the example below, version
    1.4.2 is enforced:
    ``python setup_environment.py -s 1.4.2``

Note that ``-s`` takes precedence over ``build-system-version``. If none of the two options listed above are used, the
script simply takes the latest version of the dependency manager.

To reinstall all dependencies by deleting both venvs and local poetry cache:
    ``python setup_environment.py -f``

To reinstall all dependencies by deleting venvs, local poetry cache and lock file. This option must be
used only if REALLY necessary as the installation time for the whole process will be quite long.
    ``python setup_environment.py -F``

What this script is doing?
--------------------------

The aim of this script is to automate the project installation. It creates a virtual environment, installs a
dependency manager if a configuration file is detected at project's root, configures the dependency manager to
enable package collection from private sources, installs the dependency groups following user desire.

In case ``poetry`` is detected as the project's dependency manager, two virtual environment are created. The first one
is named ``.venv`` and is located in a ``.poetry`` directory created at project's root. It is use to install
``poetry``. The second one, also named ``.venv`` is created at project's root and is controlled by ``poetry`` to
install the dependencies. This configuration is required by ``poetry``.

Future development
------------------

Given its size and its scope, this script will be refactored and transformed into a package.
"""

# ==================================================== [Imports] ==================================================== #

import argparse
import os
from pathlib import Path
import platform
import re
import shutil
import subprocess
import sys
import textwrap
import time

try:
    from packaging.markers import Marker
except:
    sys.exit("Process stopped. The `packaging` library is missing. Install with: `pip install packaging`.")
try:
    import toml
except:
    sys.exit("Process stopped. The `toml` library is missing. Install with: `pip install toml`.")

# =================================================== [Variables] =================================================== #

DEPENDENCY_MANAGER_PATHS = {
    "win32": {
        "poetry_python_executable": Path(".poetry").absolute() / ".venv" / "Scripts" / "python.exe",
        "build_sys_exec": Path(".poetry").absolute() / ".venv" / "Scripts" / "poetry.exe",
        "dep_bin_venv_path": Path(".venv").absolute() / "Scripts" / "poetry.exe",
        "python_executable": Path(".venv").absolute() / "Scripts" / "python.exe",
        "shell": True,
    },
    "linux": {
        "poetry_python_executable": Path(".poetry").absolute() / ".venv" / "bin" / "python",
        "build_sys_exec": Path(".poetry").absolute() / ".venv" / "bin" / "poetry",
        "dep_bin_venv_path": Path(".venv").absolute() / "bin" / "poetry",
        "python_executable": Path(".venv").absolute() / "bin" / "python",
        "shell": False,
    },
    "common": {
        "configuration_file": "pyproject.toml",
        "build_backend": "poetry.core.masonry.api",
        "required_venv_name": ".venv",
        "lock_file": "poetry.lock",
        "build_system_venv": Path(".poetry") / ".venv",
        "cache_folder": Path(".poetry").absolute() / ".cache",
    },
}

configuration = toml.load(DEPENDENCY_MANAGER_PATHS["common"]["configuration_file"])
STANDARD_OPTIONAL_DEPENDENCY_GROUPS = [group for group in configuration["tool"]["poetry"]["group"].keys()]

# =================================================== [Functions] =================================================== #

# Console prints ----------------------------------------------------------------------------------------------------


def print_main_header(text: str, max_length: int = 100) -> None:
    """Display main header."""

    for i in range(max_length):
        print("=", end="")
    print()
    print(text)
    for i in range(max_length):
        print("=", end="")
    print()
    print()


def print_section_header(text: str, max_length: int = 100) -> None:
    """Display a section header in the console."""

    section_header = ""
    if len(text) < max_length:
        section_header = text + " "
        for i in range(len(text), max_length):
            section_header += "-"
    else:
        section_header = text
    print(section_header)
    print()


def print_input_value(input: str, value: str, separator: str = ":", separator_position: int = 60) -> None:
    """Print input value in console."""

    if len(input) < separator_position:
        text = input
        for i in range(len(text), separator_position):
            text += " "
        text += separator + " " + value
    else:
        text = input + " " + separator + " " + value
    print(text)


def print_inputs_summary(args: object) -> None:
    """Display a summary of the inputs."""

    print(f"OS                                   : {platform.system()}")
    print(f"Python version                       : {get_python_version()}")
    print(f"Virtual environment name             : {args.venv_name}")
    print(f"run dependencies                     : {'yes' if 'run' in args.dependencies else 'no'}")
    for dependency_group in STANDARD_OPTIONAL_DEPENDENCY_GROUPS:
        print_input_value(
            f"{dependency_group} dependencies",
            "yes" if dependency_group in args.dependencies else "no",
            separator=":",
            separator_position=37,
        )
    if args.extra_dependencies:
        for dependency_group in args.extra_dependencies:
            print_input_value(
                f"{dependency_group} dependencies",
                "yes",
                separator=":",
                separator_position=37,
            )
    print(f"Dependency management system         : poetry")
    print(f"Dependency management system version : {args.build_system_version}")
    print(f"Credentials management               : {args.credentials_management_method}")
    print()


# Checks ------------------------------------------------------------------------------------------------------------


def check_virtual_environment_name(args: object) -> None:
    """Check if the virtual environment name is consistent with the build system expectations."""

    if DEPENDENCY_MANAGER_PATHS["common"]["required_venv_name"] != args.venv_name:
        old_name = args.venv_name
        args.venv_name = DEPENDENCY_MANAGER_PATHS["common"]["required_venv_name"]
        print(
            f"Warning: poetry expects the name of the virtual environment name to be"
            f" {args.venv_name}. {old_name} is replaced by {args.venv_name}."
        )


def check_python_version(args: object) -> None:
    """
    Check if the version of the current Python interpreter is consistent with the python version specifications
    from the build system.
    """

    # Initialize lower/upper versions
    lower_version, upper_version, lower_bound_symbol, upper_bound_symbol = None, None, None, None
    lower_specification, upper_specification, single_specification = None, None, None
    # Read TOML
    configuration = toml.load(DEPENDENCY_MANAGER_PATHS["common"]["configuration_file"])
    # Read Python version constraint
    python_compatibility = configuration["tool"]["poetry"]["dependencies"]["python"].replace(" ", "")
    # Split lower/upper version constraint
    if "," in python_compatibility:
        lower_specification, upper_specification = (
            python_compatibility.split(",")[0],
            python_compatibility.split(",")[1],
        )
    else:
        if python_compatibility.startswith(">"):
            lower_specification, upper_specification = python_compatibility, None
        elif python_compatibility.startswith("<"):
            lower_specification, upper_specification = None, python_compatibility
        else:
            single_specification = python_compatibility
    # Process lower constraint
    if lower_specification:
        lower_version = get_version_from_python_specification(lower_specification)
        lower_bound_symbol = get_sign_from_python_specification(lower_specification)
        marker = Marker(f"python_full_version {lower_bound_symbol} '{lower_version}'")
        if not marker.evaluate():
            raise Exception(f"Python version must be {lower_bound_symbol} {lower_version}.")
    # Process upper constraint
    if upper_specification:
        upper_version = get_version_from_python_specification(upper_specification)
        upper_bound_symbol = get_sign_from_python_specification(upper_specification)
        marker = Marker(f"python_full_version {upper_bound_symbol} '{upper_version}'")
        if not marker.evaluate():
            raise Exception(f"Python version must be {upper_bound_symbol} {upper_version}.")
    # Process single
    if single_specification:
        version = get_version_from_python_specification(single_specification)
        sign = get_sign_from_python_specification(single_specification)
        if sign != "==":
            raise Exception("Unable to interpret python version specification.")
        marker = Marker(f"python_full_version {sign} '{version}'")
        if not marker.evaluate():
            raise Exception(f"Python version must be equal to {version}.")


def check_existing_install(args: object) -> str:
    """Check if an install exists already."""

    return os.path.isdir(args.venv_name)


def check_inputs(args: object) -> None:
    """Check inputs consistency."""
    # Rework dependencies argument if all option is selected
    if "all" in args.dependencies:
        args.install_all = True
        args.dependencies = STANDARD_OPTIONAL_DEPENDENCY_GROUPS + ["run"]
    else:
        args.install_all = False
    # Check virtual environment name
    check_virtual_environment_name(args)
    # Check python version
    check_python_version(args)
    # Check if an install already exists
    args.has_install = check_existing_install(args)


# General-purpose ---------------------------------------------------------------------------------------------------


def extract_substring_between_markers(string: str, marker_1: str, marker_2: str) -> str:
    """Extract substring between two markers using find() and slice()."""

    # find() method will search the given marker and stores its index
    mk1 = string.find(marker_1) + len(marker_1)
    # find() method will search the given marker and sotres its index
    mk2 = string.find(marker_2, mk1)
    # using slicing substring will be fetched in between markers.
    return string[mk1:mk2]


def read_integers_from_string(string: str) -> list:
    """
    Scan a string and extract integers. The regex matches any digit character (0-9).
    """

    return re.findall(r"\d+", string)


def remove_scheme_from_url(url: str) -> tuple:
    """Remove the scheme part of a URL."""

    items = url.split("://")
    if len(items) == 2:
        return items[0], items[1]
    else:
        raise Exception(f"Fail to return the URL without the scheme part for {url}.")


def get_python_version() -> None:
    """Get Python version."""

    return f"{sys.version_info.major}.{sys.version_info.minor}.{sys.version_info.micro}"


def set_pip_command(
    package_version: str = "*",
    method: str = "install",
    pypi_url: str = "https://pypi.org/simple",
    private_pypi_token_name: str = "",
    no_deps: bool = False,
    python_executable: str = sys.executable,
) -> str:
    """Make pip command."""

    package_name = "poetry"

    # Add package version if specified
    if package_version != "*":
        package_name += f"=={package_version}"
    # Set command
    command = [python_executable, "-m", "pip", method, package_name]
    # Add part related to private PyPI source
    if pypi_url != "https://pypi.org/simple":
        # Check if token is available
        if private_pypi_token_name is None:
            raise Exception(f"No token specified for private PyPI server {pypi_url}.")
        # Check if the environment variable associated to the private PyPI token exists
        private_pypi_token = os.getenv(private_pypi_token_name)
        if private_pypi_token is None:
            raise Exception(f"Environment variable {private_pypi_token_name} does not exist.")
        # Split absolute URL
        url_scheme, relative_url = remove_scheme_from_url(pypi_url)
        # Update command line
        command += f" -i {url_scheme}://PAT:{private_pypi_token}@{relative_url}"
    # Add no-deps option
    if no_deps:
        command += " --no-deps"

    return command


def get_python_package(
    package_version: str = "*",
    method: str = "install",
    pypi_url: str = "https://pypi.org/simple",
    private_pypi_token_name: str = "",
    no_deps: bool = False,
    python_executable: str = sys.executable,
) -> None:
    """
    Install or download a python package using PIP.

    Parameters
    ----------
    package_version : str
        Version of the package.
    method : str
        Get method: ``download`` or ``install``
    pypi_url : str
        URL of the PyPI server.
    private_pypi_token_name : str
        Name of the environment variable holding the token to access to the PyPI.
    python_executable : str
        Path to the Python executable.

    Returns
    -------
    None
    """

    # Set pip command
    command = set_pip_command(
        package_version=package_version,
        method=method,
        pypi_url=pypi_url,
        private_pypi_token_name=private_pypi_token_name,
        no_deps=no_deps,
        python_executable=python_executable,
    )
    # Run pip command
    process = subprocess.run(
        command, check=False, shell=DEPENDENCY_MANAGER_PATHS[sys.platform]["shell"], capture_output=True, text=True
    )
    if process.returncode != 0:
        print(process.stdout)
        print(process.stderr)
        raise Exception("Command failed with error.")


def get_python_package_versions(
    pypi_url: str = "https://pypi.org/simple",
    private_pypi_token_name: str = "",
    python_executable: str = sys.executable,
) -> None:
    """Get available versions of a package."""

    # Set pip command
    command = set_pip_command(
        package_version="x",
        pypi_url=pypi_url,
        private_pypi_token_name=private_pypi_token_name,
        python_executable=python_executable,
    )
    # Run pip command
    process = subprocess.run(
        command, check=False, shell=DEPENDENCY_MANAGER_PATHS[sys.platform]["shell"], capture_output=True, text=True
    )
    if process.returncode == 0:
        print(command)
        raise Exception("This command should have failed. There is something wrong.")
    else:
        return extract_substring_between_markers(process.stderr, "(from versions:", ")").replace(" ", "").split(",")


def upgrade_pip(python_executable: str = sys.executable) -> None:
    """Upgrade to latest PIP version in the virtual environment."""

    print("Upgrade to latest pip version")
    subprocess.run(
        [python_executable, "-m", "pip", "install", "--upgrade", "pip"],
        check=True,
        shell=DEPENDENCY_MANAGER_PATHS[sys.platform]["shell"],
    )
    print()


def create_virtual_environment(args: object, venv: str = ".venv") -> None:
    """Create a virtual environment."""

    print("Create virtual environment")
    if not args.has_install or args.force_clear or args.force_clear_all:
        if sys.platform == "linux":
            subprocess.run(["sudo", "apt-get", "install", "-y", "python3-venv"], check=True)
        subprocess.run(
            [sys.executable, "-m", "venv", venv], check=True, shell=DEPENDENCY_MANAGER_PATHS[sys.platform]["shell"]
        )
    else:
        print("Skipped")
    print()


# Dependency Management System (Build System) -----------------------------------------------------------------------


def get_private_sources(configuration_file: str) -> list:
    """Get list of private sources from configuration file."""

    configuration = toml.load(configuration_file)
    try:
        private_sources = configuration["tool"]["poetry"]["source"]
    except:
        private_sources = []
    return private_sources


def get_version_from_python_specification(specification: str) -> str:
    """Read the version."""

    return ".".join(read_integers_from_string(specification))


def get_sign_from_python_specification(specification: str) -> str:
    """Read the mathematical symbol expressing a constraint on the version."""

    first_part = specification.split(".")[0]
    sign = "".join([i for i in first_part if not i.isdigit()])
    if sign == "":
        sign = "=="
    return sign


def get_build_system_version(args: object) -> str:
    """Assign build system version."""

    if not args.has_install or args.force_clear or args.force_clear_all:
        configuration = toml.load(DEPENDENCY_MANAGER_PATHS["common"]["configuration_file"])
        if "build-system-requirements" in configuration.keys():
            if (
                args.build_system_version == "*"
                and "build-system-version" in configuration["build-system-requirements"]
            ):
                args.build_system_version = configuration["build-system-requirements"]["build-system-version"]
        else:
            args.build_system_version = get_python_package_versions()[-1]


def install_build_system(args: object) -> None:
    """Install build system."""

    print("Install dependency management system.")
    if not args.has_install or args.force_clear or args.force_clear_all:
        configuration = toml.load(DEPENDENCY_MANAGER_PATHS["common"]["configuration_file"])
        if args.build_system_version == "*" and "build-system-version" in configuration["build-system-requirements"]:
            args.build_system_version = configuration["build-system-requirements"]["build-system-version"]

        create_virtual_environment(args, venv=DEPENDENCY_MANAGER_PATHS["common"]["build_system_venv"])
        upgrade_pip(python_executable=DEPENDENCY_MANAGER_PATHS[sys.platform]["poetry_python_executable"])
        get_python_package(
            args.build_system_version,
            method="install",
            python_executable=DEPENDENCY_MANAGER_PATHS[sys.platform]["poetry_python_executable"],
        )
        print()
        # Create a file symbolic link from the virtual environment (.venv) that the build system (poetry) manages
        # to the build system executable (poetry.exe) in the poetry virtual environment (.poetry/.venv).  This
        # ensures that the correct poetry is accessible when the managed virtual environment is activated
        build_system_executable = DEPENDENCY_MANAGER_PATHS[sys.platform]["build_sys_exec"]
        if os.path.exists(build_system_executable):
            if sys.platform == "win32":
                subprocess.run(
                    [
                        "powershell",
                        "-Command",
                        "New-Item",
                        "-ItemType",
                        "SymbolicLink",
                        "-Path",
                        DEPENDENCY_MANAGER_PATHS[sys.platform]["dep_bin_venv_path"],
                        "-Target",
                        build_system_executable,
                    ]
                )
            elif sys.platform == "linux":
                subprocess.run(
                    [
                        "ln",
                        "-sf",
                        build_system_executable,
                        DEPENDENCY_MANAGER_PATHS[sys.platform]["dep_bin_venv_path"],
                    ]
                )
        return
    print("Skipped")
    print()


def configure_build_system(args: object) -> None:
    """Configure the build system to enable connection to private sources."""

    print("Configure dependency management system.")
    if not args.has_install or args.force_clear or args.force_clear_all:
        configure_poetry(
            DEPENDENCY_MANAGER_PATHS["common"]["build_system_venv"],
            args.credentials_management_method,
        )
        print()
        return
    print("Skipped")
    print()


def configure_poetry(venv_name: str, credentials_management_method: str) -> None:
    """Configure Poetry."""
    poetry_executable = DEPENDENCY_MANAGER_PATHS[sys.platform]["dep_bin_venv_path"]
    # Get list of private sources
    private_sources = get_private_sources("pyproject.toml")
    # Turn-on in-project
    subprocess.run([poetry_executable, "config", "virtualenvs.create", "false", "--local"], check=True)
    # Turn-off virtual environment creation
    subprocess.run([poetry_executable, "config", "virtualenvs.in-project", "true", "--local"], check=True)
    # Turn-on poetry cache
    subprocess.run(
        [
            poetry_executable,
            "config",
            "cache-dir",
            DEPENDENCY_MANAGER_PATHS["common"]["cache_folder"],
            "--local",
        ],
        check=True,
    )
    # Turn-on in-project
    subprocess.run([poetry_executable, "config", "virtualenvs.create", "false", "--local"], check=True)
    # Declare credentials for private sources
    for source in private_sources:
        print(f"Declare credentials for {source['name']}")
        if source["name"].lower() == "pypi":
            continue
        elif source["url"] == "https://pkgs.dev.azure.com/pyansys/_packaging/pyansys/pypi/simple/":
            token = os.environ["PYANSYS_PRIVATE_PYPI_PAT"]
        elif source["url"] == "https://pkgs.dev.azure.com/pyansys/_packaging/ansys-solutions/pypi/simple/":
            token = os.environ["SOLUTIONS_PRIVATE_PYPI_PAT"]
        else:
            raise Exception(f"Unknown private source {source['name']} with url {source['url']}.")
        # Store credentials
        if credentials_management_method == "keyring":
            # Declare source URL
            command_line = [poetry_executable, "config", f"repositories.{source['name']}", source["url"], "--local"]
            subprocess.run(command_line, check=True)
            # Declare source credentials
            command_line = [poetry_executable, "config", f"http-basic.{source['name']}", "PAT", token, "--local"]
            subprocess.run(command_line, check=True)
        elif credentials_management_method == "environment-variables":
            # Format source name to comply with Poetry environment variable syntax
            source_name = source["name"].upper().replace("-", "_")
            # Create Poetry environment variable
            os.environ[f"POETRY_HTTP_BASIC_{source_name}_USERNAME"] = "PAT"
            os.environ[f"POETRY_HTTP_BASIC_{source_name}_PASSWORD"] = token


def check_dependency_group(dependency_group: str, configuration: str) -> bool:
    """Return True if the dependency group is available in the configuration file."""

    try:
        configuration["tool"]["poetry"]["group"][dependency_group]
        return True
    except:
        return False


# Specifics ---------------------------------------------------------------------------------------------------------


def parser() -> None:
    """Parse command line arguments."""

    # Code Name
    program_name = "Setup Environment Utility"
    # Code description
    program_description = "A Python script to automate the setup of the Python ecosystem of a project."
    # Create top-level parser
    parser = argparse.ArgumentParser(
        prog=program_name,
        usage=None,
        prefix_chars="-",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=textwrap.dedent(program_description),
    )
    parser._action_groups.pop()
    # Definition of the group of arguments
    optional_inputs = parser.add_argument_group("Optional arguments")
    # Optional parameters
    optional_inputs.add_argument(
        "-d",
        "--dependencies",
        type=str,
        nargs="+",
        help="List of dependency groups to be installed separated by space. Example: run tests doc style build",
        default=[],
        choices=STANDARD_OPTIONAL_DEPENDENCY_GROUPS + ["run", "all"],
        required=False,
    )
    optional_inputs.add_argument(
        "-x",
        "--extra-dependencies",
        type=str,
        nargs="+",
        help="List of extra dependency groups to be installed separated by space.",
        default=[],
        required=False,
    )
    optional_inputs.add_argument(
        "-s",
        "--build-system-version",
        type=str,
        help="Build system version",
        default="*",
        required=False,
    )
    optional_inputs.add_argument(
        "-c",
        "--credentials-management-method",
        type=str,
        help="Method to store private sources credentials.",
        default="keyring",
        choices=["keyring", "environment-variables"],
        required=False,
    )
    optional_inputs.add_argument(
        "-e",
        "--venv-name",
        type=str,
        help="Name of the virtual environment. Must be '.venv' when using poetry",
        default=".venv",
        required=False,
    )
    optional_inputs.add_argument(
        "-F",
        "--force-clear-all",
        help="Clean-up the workspace. Delete existing .venv, .poetry\.venv, .poetry\.cache and poetry.lock.",
        action="store_true",
        required=False,
    )
    optional_inputs.add_argument(
        "-f",
        "--force-clear",
        help="Clean-up the workspace. Delete existing .venv., .poetry\.venv and .poetry\.cache.",
        action="store_true",
        required=False,
    )
    optional_inputs.add_argument(
        "-v",
        "--verbose",
        help="Activate verbose mode.",
        action="store_true",
        required=False,
    )

    return parser.parse_args()


def clear_workspace(args: object) -> None:
    """Remove residual items form previous installation (like venv directory, lock file ...)."""

    print("Clear workspace")
    if args.force_clear or args.force_clear_all:
        # Remove virtual environment
        if os.path.isdir(args.venv_name):
            print(f"Delete existing virtual environment {args.venv_name}.")
            shutil.rmtree(args.venv_name)
        # Remove poetry virtual environment
        if os.path.isdir(DEPENDENCY_MANAGER_PATHS["common"]["build_system_venv"]):
            print("Delete existing poetry virtual environment")
            shutil.rmtree(DEPENDENCY_MANAGER_PATHS["common"]["build_system_venv"])
        # Remove poetry cache
        if os.path.isdir(DEPENDENCY_MANAGER_PATHS["common"]["cache_folder"]):
            print("Delete existing poetry cache")
            shutil.rmtree(DEPENDENCY_MANAGER_PATHS["common"]["cache_folder"])
        if args.force_clear_all:
            # Remove lock file
            if DEPENDENCY_MANAGER_PATHS["common"]["lock_file"]:
                if os.path.isfile(DEPENDENCY_MANAGER_PATHS["common"]["lock_file"]):
                    print("Delete existing poetry lock file")
                    os.remove(DEPENDENCY_MANAGER_PATHS["common"]["lock_file"])
    else:
        print("Skipped")
    print()


def install_production_dependencies(args: object) -> None:
    """Install the package (mandatory requirements only)."""

    print("Install production dependencies")
    if "run" in args.dependencies:
        subprocess.run(
            [
                DEPENDENCY_MANAGER_PATHS[sys.platform]["build_sys_exec"],
                "install",
                "-vv",
            ],
            check=True,
            shell=DEPENDENCY_MANAGER_PATHS[sys.platform]["shell"],
        )
        print()
        return
    print("Skipped")
    print()


def install_optional_dependencies(args: object) -> None:
    """Install optional requirements (doc, tests, build or style)."""

    # Load configuration file
    configuration = toml.load(DEPENDENCY_MANAGER_PATHS["common"]["configuration_file"])
    # Install standard optional dependency groups
    for dependency_group in STANDARD_OPTIONAL_DEPENDENCY_GROUPS:
        print(f"Install {dependency_group} dependencies")
        if dependency_group in args.dependencies:
            # Install dependency group in the configuration file of the build system
            has_dependency_group = check_dependency_group(dependency_group, configuration)
            if has_dependency_group:
                print("Installing from build system configuration file.")
                subprocess.run(
                    [
                        DEPENDENCY_MANAGER_PATHS[sys.platform]["build_sys_exec"],
                        "install",
                        "--only",
                        dependency_group,
                        "-vv",
                    ],
                    check=True,
                    shell=DEPENDENCY_MANAGER_PATHS[sys.platform]["shell"],
                )
                print()
                continue
            # Alternatively search dependency group in the requirements folder
            requirements_file = os.path.join("requirements", f"requirements_{dependency_group}.txt")
            if os.path.exists(requirements_file):
                print("Installing from requirements file.")
                subprocess.run(
                    [
                        DEPENDENCY_MANAGER_PATHS[sys.platform]["poetry_python_executable"],
                        "-m",
                        "poetry",
                        "run",
                        "pip",
                        "install",
                        "-r",
                        requirements_file,
                        "--no-warn-conflicts",
                    ],
                    check=True,
                    shell=DEPENDENCY_MANAGER_PATHS[sys.platform]["shell"],
                )
                print()
                continue
        print("Skipped")
        print()
    # Install extra optional dependency groups
    if args.extra_dependencies:
        for dependency_group in args.extra_dependencies:
            print(f"Install {dependency_group} dependencies")
            has_dependency_group = check_dependency_group(dependency_group, configuration)
            if has_dependency_group:
                subprocess.run(
                    [
                        DEPENDENCY_MANAGER_PATHS[sys.platform]["build_sys_exec"],
                        "install",
                        "--only",
                        dependency_group,
                        "-vv",
                    ],
                    check=True,
                    shell=DEPENDENCY_MANAGER_PATHS[sys.platform]["shell"],
                )
            else:
                print(f"No dependency group named {dependency_group}.")
                print("Skipped")
            print()


def install_dotnet_linux_dependencies():
    print("Install dotnet dependencies")
    if sys.platform == "linux":
        subprocess.run(
            'set -xe \
            && wget https://dot.net/v1/dotnet-install.sh \
            && chmod +x dotnet-install.sh \
            && ./dotnet-install.sh --install-dir /home/$USER/.dotnet --version 3.1.0 --runtime aspnetcore \
            && grep -qxF "DOTNET_ROOT=/home/$USER/.dotnet" /home/$USER/.bash_profile \
            || echo DOTNET_ROOT=/home/$USER/.dotnet >> /home/$USER/.bash_profile \
            && grep -qxF "PATH=\$PATH:/home/$USER/.dotnet" /home/$USER/.bash_profile \
            || echo "PATH=\$PATH:/home/$USER/.dotnet" >> /home/$USER/.bash_profile \
            && echo "\nsource /home/$USER/.bash_profile" >> ./.venv/bin/activate \
            && rm dotnet-install.sh',
            check=True,
            shell=True,
        )


def main() -> None:
    """Sequence of operations leading to the complete Python ecosystem."""

    # Start timer
    time_on = time.time()
    # Get current working directory
    working_directory = os.getcwd()
    # Get installation directory
    install_directory = Path(__file__).parent.absolute()
    # Move to install directory
    os.chdir(install_directory)

    # Initialize session ----------------------------------------------------------------------------------------------

    # Read command line inputs
    args = parser()

    # Check inputs consistency
    check_inputs(args)

    # Update build system version
    get_build_system_version(args)

    # Display header
    print_main_header("Setup Environment")

    # Display inputs summary
    print_inputs_summary(args)

    # Clear workspace -------------------------------------------------------------------------------------------------

    print_section_header("Clear workspace", max_length=100)

    clear_workspace(args)

    # Setup virtual environment ---------------------------------------------------------------------------------------

    print_section_header("Setup virtual environment", max_length=100)

    create_virtual_environment(args)

    upgrade_pip(python_executable=DEPENDENCY_MANAGER_PATHS[sys.platform]["python_executable"])

    # Setup dependency management system ------------------------------------------------------------------------------

    print_section_header("Setup dependency management system", max_length=100)

    install_build_system(args)

    configure_build_system(args)

    # Install dependencies --------------------------------------------------------------------------------------------

    print_section_header("Install dependencies", max_length=100)

    install_production_dependencies(args)

    install_optional_dependencies(args)

    install_dotnet_linux_dependencies()

    # Back to current working directory
    os.chdir(working_directory)

    # Compute execution time
    elapsed_time = (time.time() - time_on) / 60  # in minutes

    print("You are all set!")
    print("Navigate to project root and activate your environment with one these commands:")
    print(rf"   - For Windows CMD       : {args.venv_name}\Scripts\activate.bat")
    print(rf"   - For Windows Powershell: {args.venv_name}\Scripts\Activate.ps1")
    print(rf"   - For Ubuntu            : source {args.venv_name}/bin/activate")
    print("Enjoy!")
    print()
    print("Execution time: %.1f minutes." % (elapsed_time))
    print()


# =================================================== [Execution] =================================================== #

if __name__ == "__main__":
    main()
